
\ Callist2.f 
\ Каллисто-2, черновик 21
\ Авторское право (c) Васильев Илья Владимирович, 6 декабря 2018, Москва
\ Перед вами черновик языка программирования для "Электроники МК-161", развивающего идеи Каллисто Классик

\ Каллисто-2 распространяется под свободной лицензией GNU GPL версии 3 или более поздней.
\ Текст лицензии GPL на русском и английском языках располагается в архиве дистрибутива в каталоге gpl.
\ Каллисто-2, а также основаные на ней работы, запрещается распространять без исходного текста и текста лицензии.
\ Распространение приложений, написанных на Каллисто-2, и использование примеров кода на Каллисто-2,
\ размещённых в Руководстве, разрешается без ограничений.
\
\ Если вы несогласны с правами, которые предоставляет свободное программное обеспечение
\ и обязанностями, которые оно налагает, удалите Каллисто-2 со своих носителей.

\ Исходный текст предназначен для SP-Forth 4.20

\ == Регистры МК-161 ==
\ R0           Временный регистр обработчика
\ R2 RP        Указатель целочисленного стека возвратов, растёт вниз от 5090
\ R3 SP        Указатель десятичного стека: растёт вниз от 998
\ R5           Временный регистр обработчика
\ R6 RI, IP    Указатель инструкций, указывает на токен в шитом коде
\ R7 W, WP     Указатель слов Форта, содержит CFA исполняемого слова
\              временный регистр обработчика
\ R8 JP        Указатель/регистр передачи управления NEXT,
\              временный регистр обработчика
\ R9 =NEXT     Адрес точки NEXT
\
\ R19 = 9007/9008 (DARK/LIGHT) номер регистра прокрутки для текущего цвета фона
\ R20          Начало десятичного словаря
\
\ R9042 =9500  Индексный регистр памяти программ
\              Содержит адрес таблицы свёрток

ONLY FORTH DEFINITIONS
VOCABULARY VMETA        \ Метаязык, образы памяти МК-161, code start-code
VOCABULARY VMK          \ Язык МК, управляющие и вспомогательные слова, end-code
VOCABULARY VTARGET      \ Словарь, дублирующий целевой и метки в целевых пространствах

\ Для отладки
ALSO VMETA  DEFINITIONS
: .s ( -- )  ." Stack: " DEPTH IF DEPTH 0 DO I PICK . LOOP ELSE ." empty" THEN CR ;  \ Печать инструментального стека
: .HH ( b -- )  BASE @ SWAP HEX 0 <# # # #> TYPE SPACE BASE ! ;                      \ Шестнадцатеричная печать байта
: .8 ( a -- ) 8 0 DO DUP C@ .HH 1+ LOOP DROP ;                                       \ Простой инстументальный дамп

\ --------------------------------------
\ Области памяти МК-161 и их сохранение.
\ --------------------------------------

CREATE prgm  10000 ALLOT         \ Память программ
prgm 10000 255 FILL              \ Заполняем FF

: .prg ( a -- ) prgm + DUP .8 8 + .8 ;

CREATE regb  8168 ALLOT          \ Регистровая память (байты)
regb 8168 ERASE                  \ Обнуляем

CREATE regd  8000 ALLOT          \ Регистровая память (десятичные регистры)
regd 8000 ERASE                  \ Обнуляем

: save.mkp ( -- )                \ Сохранить память программ в файл Callist2.mkp
  HERE 0 C,                      \ Номер начальной страницы
  prgm 10000 0 DO DUP C@ C, 1 C, 1+ LOOP DROP       \ Создать образ
  S" Callist2.mkp" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKP"
  SWAP 20001 2 PICK
  WRITE-FILE  ABORT" Can't write MKP"
  CLOSE-FILE  ABORT" Can't close MKP"
  -20001 ALLOT                                      \ Убрать образ из памяти
;

: save.mkb ( -- )                \ Сохранить байтовые регистры в файл Callist2.mkb
  S" Callist2.mkb" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKB"
  1000 regb + 4096 2 PICK
  WRITE-FILE  ABORT" Can't write MKB"
  CLOSE-FILE  ABORT" Can't close MKB"
;

: save.mkd ( -- )                \ Сохранить десятичные регистры в файл Callist2.mkd
  S" Callist2.mkd" W/O ( BIN )
  CREATE-FILE ABORT" Can't create MKD"
  regd 8000 2 PICK
  WRITE-FILE  ABORT" Can't write MKD"
  CLOSE-FILE  ABORT" Can't close MKD"
;

\ ----------------------
\ Вспомогательные слова.
\ ----------------------
: 4*  ( n1 -- n2 )  2* 2* ;                           \ Умножение на 4
: 8*  ( n1 -- n2 )  2* 2* 2* ;                        \ Умножение на 8
: 16* ( n1 -- n2 )  8* 2* ;                           \ Умножение на 16
: [+] + ;  : [-] - ;  : [1-] 1- ;  : [1+] 1+ ;        \ Сохраним операции с инструментальным стеком данных, как [+] и [-]

: WBin@ ( адрес -- n )                                \ Считать 2-х байтовое целое без знака в двоичном формате ЭКВМ
    DUP C@ 16* 16* SWAP 1+ C@ + ;
: WBin! ( n адрес -- )  OVER 256 / OVER C!            \ Записать 2-х байтовое целое без знака в двоичном формате ЭКВМ
    1+ SWAP 256 MOD SWAP C! ;
: WZero! ( адрес -- ) 0 OVER C!  1+ 0 SWAP C! ;       \ Обнулить два байта
: C>BCD ( c -- n ) 10 /MOD 16* + ;
: BCD>C ( n -- c ) 16 /MOD 10 * + ;
: WBCD@ ( адрес -- n )
    DUP C@ BCD>C 100 *  SWAP 1+ C@ BCD>C + ;          \ Считать 2-х байтовое целое число без знака в bcd-формате ЭКВМ
: WBCD! ( n адрес -- ) OVER 100 / C>BCD OVER C!       \ Записать 2-х байтовое целое число без знака в bcd-формате ЭКВМ
    1+ SWAP 100 MOD C>BCD SWAP C! ;

: getnumber ( "число" -- n )                          \ Считать следующее слово из входного потока, как число (нужен ли?)
    0 0 BL WORD COUNT >NUMBER 2DROP DROP ;

: ShiftDigits ( a -- )                                \ Сдвинуть мантиссу вправо, освободив место для новой цифры слева
  7 OVER + DUP C@ 1+ SWAP C!                          \ Увеличить порядок на 1
  0                                                   \ Вставить цифру 0
  6 0 DO                                              \ Цикл выполнится 6 раз
  ( a цифра )
    16* SWAP 1+ SWAP    ( a+1 16*цифра )
    OVER C@ 16 /MOD     ( a+1 16*цифра low high )
    ROT +               ( a+1 low newbyte )
    2 PICK C!           ( a+1 low )
  LOOP
  2DROP ;

: StoreDigit ( a цифра -- )                         \ Добавить старшую десятичную цифру к числу во внутреннем формате МК-161
    16* SWAP 1+ DUP C@  ( d*16 a+1 byte )
    ROT + SWAP C! ;

: dec! ( n a -- )                                   \ Записать целое n во внутреннем формате МК-161
  DUP 8 ERASE                                       \ Обнулим 8-байтный буфер
  SWAP DUP 0< IF 0x80 2 PICK C! NEGATE THEN         \ Запишем знак числа
  ( a n )
  BEGIN 10 /MOD        ( a цифра n' )
    SWAP 2 PICK SWAP   ( a n' a цифра )
    StoreDigit         ( a n' )
    DUP
  WHILE OVER ShiftDigits
  REPEAT
  2DROP ;

\ ------------------------------------
\ Метки для ссылок вперёд. Объявления.
\ ------------------------------------
ONLY FORTH  ALSO VMETA  ALSO VMK  DEFINITIONS
1 VALUE lc                            \ Счётчик меток начинается с метки 1
0 VALUE LabelLast                     \ Метка, определяющая контекст
50 CONSTANT LBLMAX                    \ Максимальное число локальных меток
CREATE LTABLE  LBLMAX 1+ 2*  ALLOT
50 CONSTANT FRMAX                     \ Максимальное число ссылок вперёд
CREATE FRTABLE  FRMAX 4* ALLOT        \ ByteBin метка,атрибут, WBCD адрес-вхождения

LTABLE LBLMAX 1+ 2* ERASE             \ FRCLR: Инициализация таблиц ссылок вперёд
FRTABLE FRMAX 4* ERASE

: LabelClear ( -- ) 0 TO LabelLast ;  \ Очистка контекста метки

: FRCHK                               \ Проверка неразрешённых ссылок вперёд
    FRMAX 0 DO I 4* FRTABLE + C@ ABORT" FRCHK: Unresolved forward" LOOP ;


\ ---------------------------------------------------------
\ X-компилятор исходного текста Каллисто-2 в память МК-161.
\ ---------------------------------------------------------
ONLY FORTH  ALSO VMETA  DEFINITIONS
: TCREATE ( "имя" -- )                              \ Создать слово в дублирующем словаре
    ALSO VTARGET  DEFINITIONS  CREATE  PREVIOUS ;

1000 VALUE bc                                       \ Номер текущего байтового регистра
: bhere ( -- адрес ) bc regb + ;                    \ Инструментальный адрес текущего байтового регистра
: ballot ( n -- ) bc + TO bc ;                      \ Зарезервировать n байтовых регистров
: b, ( c -- ) bhere C! 1 ballot ;                   \ Скомпилировать байт c в байтовую память
: bw, ( n -- ) 256 /MOD b, b, ;                     \ Скомпилировать целое n в байтовую память
: bxt, ( xt -- ) 100 /MOD b, b, ;                   \ Скомпилировать выполнимый токен xt в байтовую память
: .borg ( "число" -- ) getnumber TO bc ;            \ Сделать байтовый регистр aaaa текущим
: b: ( "имя" -- ) TCREATE bc , DOES> ( -- bc ) @ ;  \ Определить метку в байтовой памяти
: breg ( "имя" -- ) b: 1 ballot ;                   \ Дать имя очередному байтовому регистру
: btext" ( "text"" -- )                             \ Скомпилировать строку литер в байтовую память
    34 WORD COUNT SWAP bhere 2 PICK MOVE ballot ;

20 VALUE dc                                         \ Распределение десятичных регистров начинается с R20
: d! ( n a -- ) 8* regd + dec! ;                    \ Присвоить целое значение десятичному регистру
: dallot ( n -- ) dc + TO dc ;                      \ Зарезервировать n десятичных регистров, добавить проверку на конфликт с R100
: d, ( n -- ) dc d! 1 dallot ;                      \ Скомпилировать n на вершину десятичного словаря
: .dorg ( "число" -- ) getnumber TO dc ;            \ Сделать десятичный регистр aaaa текущим
: d: ( "имя" -- ) TCREATE dc , DOES> ( -- dc ) @ ;  \ Определить метку в десятичной памяти
: dreg ( "имя" -- ) d: 1 dallot ;                   \ Дать имя очередному десятичному регистру
: dconst ( "имя" n -- ) d: d, ;                     \ Дать имя и значение очередному десятичному регистру

100 VALUE ddc                                       \ Словарь FORTH начинается с R100
: ddc++ ( -- ) ddc 1 + TO ddc ;                     \ Увеличить ddc на 1
: ddc@ ( -- a ) ddc 8* regd + ;                     \ Дать инструментальный адрес текущего регистра-заголовка
: dheader! ( hash1 hash2 a -- )                     \ Записать заголовок (CFA и хэш) в десятичный регистр (по инструментальному адресу)
    0 OVER C! 1+                \ 00 Знак 0 - положительное
    bc OVER WBCD! 2+            \ 01-02 Мантисса - CFA=bc
    ROT OVER WBCD! 2+           \ 03-04 Мантисса - hash1
    SWAP OVER WBCD! 2+          \ 05-06 Мантисса - hash2
    3 SWAP C!                   \ 07 Порядок - 03
;
: dheader, ( hash1 hash2 -- )  ddc@ dheader! ddc++ ; \ Скомпилировать десятичный заголовок в целевой список слов FORTH

0 VALUE pc                                          \ Счётчик команд
: p! ( b a -- )  prgm + C! ;                        \ Запись шага b по адресу a в память программ
: phere ( -- адрес )  pc prgm + ;                   \ Инструментальный адрес текущего места в памяти программ
: pallot ( n -- )  pc + TO pc ;                     \ Зарезервировать n шагов в памяти программ
: p, ( c -- )  phere C! 1 pallot ;                  \ Скомпилировать шаг c в память программ
: pw, ( n -- ) 256 /MOD p, p, ;                     \ Скомпилировать целое n в память программ
: pbcd, ( n -- ) 100 MOD C>BCD p, ;                 \ Скомпилировать в шаг младшую половину n, как число bcd
: pbcd2, ( n -- ) phere WBCD! 2 pallot ;            \ Скомпилировать в два шага bcd-число n
: pbcd!  ( n a -- ) prgm +  SWAP 100 MOD C>BCD SWAP C! ;
: pbcd2! ( n a -- ) prgm +  WBCD! ;                 \ Записать n в два шага по адресу a, как число bcd
: ptext" ( "text"" -- )                             \ Скомпилировать строку литер в память программ
    34 WORD COUNT SWAP phere 2 PICK MOVE pallot ;
: p: ( "имя" -- )  TCREATE pc , DOES> ( -- pc ) @ ; \ Определить метку в памяти программ
: .org ( "число" -- )  getnumber TO pc ;            \ Сделать шаг по адресу aaaa текущим

0 VALUE sc                                          \ Номер свободной свёртки
: ps, ( pc -- ) 9500 prgm +                         \ Добавить адрес обработчика pc в таблицу свёрток
   sc 2* +  SWAP 256 /MOD
   2 PICK C!  SWAP 1+ C!  sc 1+ TO sc ;


\ -------------------------------------------------
\ X-компилятор входного языка "Электроники МК-161".
\ -------------------------------------------------
ONLY FORTH  ALSO VMETA  ALSO VMK  DEFINITIONS
: 1mi ( cop "мнемоника" -- )  CREATE C, DOES> ( -- ) C@ p, ;            \ Одношаговая команда
: 2mi ( cop "мнемоника" -- )  CREATE C, DOES> ( n -- ) C@ p,  pbcd, ;   \ Двухшаговая команда

: end-code ( -- )  PREVIOUS PREVIOUS ;              \ Завершить x-определение примитива

: ?pairs ( n1 n2 -- )                               \ Всё ли в порядке с операторными скобками?
    - ABORT" Conditionals aren't paired" ;
: ?far ( цель откуда -- флаг )                      \ Проверить необходимость межстраничного перехода
    1+ 100 /  SWAP 100 /  = 0= ;
: ?reg61 ( регистр -- флаг )  16 < ;                \ Проверить, идёт ли обращение к регистрам МК-61
: ?regPP ( регистр -- флаг )  99 > ;                \ Проверить, идёт ли обращение к регистрам выше R99

ONLY FORTH  ALSO VMETA  DEFINITIONS
: start-code ( -- )                                 \ Начать компилировать код на ЯМК
     ALSO VTARGET  DEFINITIONS  ALSO VMK ;

: code ( "имя" -- )                                 \ Начать x-определение безымянного примитива
    TCREATE bc ,  sc b,  pc ps,  start-code  DOES> ( -- ) @ bxt, ;

: #code ( "имя" hash1 hash2 -- )                    \ Начать x-определение примитива с заголовком
    dheader, TCREATE bc ,  sc b,  pc ps,  start-code  DOES> ( -- ) @ bxt, ;

: :: ( "имя" -- )                                   \ Начать x-определение высокого уровня
    TCREATE bc ,  0 b,  ALSO VMETA  ALSO VTARGET
    DOES> ( -- ) @ bxt, ;                           \ Сделать ;- для оптимизации хвоста
: ;; ( -- )  PREVIOUS PREVIOUS  ;                   \ Может компилировать exit, но пока пишем exit вручную


\ ---------------------------------------
\ Умные переходы и обращения к регистрам.
\ ---------------------------------------
ONLY FORTH  ALSO VMETA  ALSO VMK  DEFINITIONS
  7 CONSTANT x≠0                                    \ Коды условий для структур управления
  9 CONSTANT x≥0
0xC CONSTANT x<0
0xE CONSTANT x=0

\ Умные переходы
: БП   ( адрес -- )  DUP pc ?far IF 0xF1 p, pbcd2, ELSE 0x51 p, pbcd, THEN LabelClear ;
: ПП   ( адрес -- )  DUP pc ?far IF 0xF3 p, pbcd2, ELSE 0x53 p, pbcd, THEN LabelClear ;
: Fx≠0 ( адрес -- )  DUP pc ?far IF 0xF7 p, pbcd2, ELSE 0x57 p, pbcd, THEN LabelClear ;
: FL2  ( адрес -- )  DUP pc ?far IF 0xF8 p, pbcd2, ELSE 0x58 p, pbcd, THEN LabelClear ;
: Fx≥0 ( адрес -- )  DUP pc ?far IF 0xF9 p, pbcd2, ELSE 0x59 p, pbcd, THEN LabelClear ;
: FL3  ( адрес -- )  DUP pc ?far IF 0xFA p, pbcd2, ELSE 0x5A p, pbcd, THEN LabelClear ;
: FL1  ( адрес -- )  DUP pc ?far IF 0xFB p, pbcd2, ELSE 0x5B p, pbcd, THEN LabelClear ;
: Fx<0 ( адрес -- )  DUP pc ?far IF 0xFC p, pbcd2, ELSE 0x5C p, pbcd, THEN LabelClear ;
: FL0  ( адрес -- )  DUP pc ?far IF 0xFD p, pbcd2, ELSE 0x5D p, pbcd, THEN LabelClear ;
: Fx=0 ( адрес -- )  DUP pc ?far IF 0xFE p, pbcd2, ELSE 0x5E p, pbcd, THEN LabelClear ;

\ Структуры управления, можно упростить реализацию с помощью умных переходов
: begin ( -- цель 101 )  pc 101 ;                   \ Отметить, куда будет переход назад, начало цикла
: until ( цель 101 условие -- ) SWAP 101 ?pairs     \ Условный переход назад, цикл "до"
    OVER pc ?far IF 0xF0 + p, pbcd2, ELSE 0x50 + p, pbcd, THEN ;
: again ( цель 101 -- ) 101 ?pairs                  \ Безусловный переход назад, бесконечный цикл
    DUP pc ?far IF 0xF1 p, pbcd2, ELSE 0x51 p, pbcd, THEN ;
: fl0 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R0 назад, begin..fl0 = FL0/PFL0
    DUP pc ?far IF 0xFD p, pbcd2, ELSE 0x5D p, pbcd, THEN ;
: fl1 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R1 назад, begin..fl1 = FL1/PFL1
    DUP pc ?far IF 0xFB p, pbcd2, ELSE 0x5B p, pbcd, THEN ;
: fl2 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R2 назад, begin..fl2 = FL2/PFL2
    DUP pc ?far IF 0xF8 p, pbcd2, ELSE 0x58 p, pbcd, THEN ;
: fl3 ( цель 101 -- ) 101 ?pairs                    \ Цикл по R3 назад, begin..fl3 = FL3/PFL3
    DUP pc ?far IF 0xFA p, pbcd2, ELSE 0x5A p, pbcd, THEN ;

: if ( условие -- откуда 102 )  0xF0 + p,           \ Условный переход вперёд
    pc 0 p, 0 p, 102 ;                              \ (можно добавить ifs для внутристраничного перехода)
: else ( откуда 102 -- откуда 102 ) 102 ?pairs      \ Вторая часть ветвления if
    pc 3 + SWAP pbcd2!  0xF1 p,                     \ Конец условного и начало безусловного переходов
    pc 0 p, 0 p, 102 ;
: then ( откуда 102 -- )  102 ?pairs                \ Конец условного перехода вперёд
    pc SWAP pbcd2! ;
: tails ( условие -- откуда 102 )  0xF3 p,          \ Вызов подпрограммы: РПП вперёд
    pc 0 p, 0 p, 102 ;                              \ Хвостовая рекурсия: tails .. then

\ Умный доступ к регистрам МК-161
: П ( регистр -- )                                  \ Запись в произвольный регистр МК-161
    DUP ?reg61 IF 0x40 + p,
    ELSE DUP ?regPP IF 0xF4 p, pbcd2, ELSE 0x4F p, pbcd, THEN
    THEN
;

: ИП ( регистр -- )                                 \ Чтение произвольного регистра МК-161
    DUP ?reg61 IF 0x60 + p,
    ELSE DUP ?regPP IF 0xF6 p, pbcd2, ELSE 0x6F p, pbcd, THEN
    THEN
;

: pnum, ( число -- ) -1 SWAP                        \ Скомпилировать целое число в команды языка МК-161
    BEGIN 10 /MOD DUP 0= UNTIL DROP                 \ Сперва кладём все цифры в стек
    BEGIN p, DUP 0< UNTIL DROP                      \ Потом компилируем их в нужном порядке
    LabelClear
;

\ ------------------------------------
\ Метки для ссылок вперёд. Реализация.
\ ------------------------------------
: FRRES4 ( адрес -- )                 \ Атрибут 04: четырёхшаговый ввод числа, можно добавить проверку, достаточно ли 3 шагов
    pc SWAP TO pc
    DUP  1000 /MOD p,  100 /MOD p,  10 /MOD p, p,  TO pc ;

: FRRES3 ( адрес -- )                 \ Атрибут 03: трёхшаговый ввод числа
    pc 999 > ABORT" FRRES3: Constant is too big"
    pc SWAP TO pc
    DUP  100 /MOD p,  10 /MOD p, p,  TO pc ;

: FRRES2 ( адрес -- )                 \ Атрибут 02: трёхшаговый переход вроде РБП, можно добавить проверку на внутристраничный переход
    pc SWAP 1+ pbcd2! ;

: FRRES1 ( адрес -- )                 \ Атрибут 01: двухшаговый переход вроде БП
    pc OVER ?far ABORT" FRRES1: Forward reference is out of page" 
    pc SWAP 1+ pbcd! ;
    
: FRRES  ( -- )                       \ Разрешение контекстной ссылки вперёд
    FRMAX 0 DO  I 4* FRTABLE + DUP C@ LabelLast =
      IF ( вхождение )
        DUP 1+ C@                     \ Считываем атрибут
        OVER WZero!                   \ Обнуляем поля атрибута и метки
        SWAP 2+ WBCD@ SWAP            \ Получаем адрес ссылки вперёд
        DUP 4 = IF DROP FRRES4 ELSE   \ Вызываем нужный обработчик по атрибуту
        DUP 3 = IF DROP FRRES3 ELSE
        DUP 2 = IF DROP FRRES2 ELSE
        1 = IF FRRES1 THEN THEN THEN THEN
      ELSE DROP THEN
    LOOP ;

: FRADD  ( атрибут -- )               \ Добавление контекстной ссылки вперёд в таблицу
    FRMAX 1+ 0
    DO FRMAX I = ABORT" FRADD: Too many forward references"
      I 4* FRTABLE + DUP C@ 0=
      IF LabelLast OVER C!            \ Записываем метку
         pc OVER 2+ WBCD!             \ Записываем адрес
         1+ C!                        \ Записываем атрибут
         LabelClear LEAVE ELSE DROP THEN
    LOOP ;

: Label ( "метка" -- )                \ Определить метку
    CREATE  lc DUP ,  1+ DUP TO lc  LBLMAX > ABORT" Label: Too many labels"
    DOES> ( -- значение | 0 )  @ DUP TO LabelLast  2* LTABLE + WBCD@ ;

: &: ( 0 -- )                         \ Установить метку на текущее место в памяти программ МК-161
    ABORT" &:: Duplicate definition"
    FRRES  pc  LabelLast 2* LTABLE + WBCD!  LabelClear ;

: &! ( n 0 -- )                       \ Присвоить метке значение n
    ABORT" &=: Duplicate definition"
    pc SWAP TO pc
    FRRES pc LabelLast 2* LTABLE + WBCD!  LabelClear  TO pc ;

\ Возможно слияние реализации с "умными" командами для случая перехода назад
\ Ещё можно продумать ссылку на регистры, определяемые позднее
\ Также надо продумать смещения, включая отрицательные

: &2mi ( cop "мнемоника" -- )  CREATE C, DOES> ( a -- ) 1 FRADD  C@ p,  pbcd, ;   \ Двухшаговая команда со ссылкой вперёд
: &3mi ( cop "мнемоника" -- )  CREATE C, DOES> ( a -- ) 2 FRADD  C@ p,  pbcd2, ;  \ Трёхшаговая команда со ссылкой вперёд

: &&&,  ( n -- )
    DUP 999 > ABORT" &&&,: Constant is too big"
    3 FRADD  100 /MOD p,  10 /MOD p, p, ;
: &&&&,  ( n -- )  4 FRADD 1000 /MOD p,  100 /MOD p,  10 /MOD p, p, ;    \ Можно добавить проверку, достаточно ли 3 шагов

\ -----------------------
\ Таблица кодов операций.
\ -----------------------
   0 1mi 0         1 1mi 1          2 1mi 2          3 1mi 3         4 1mi 4         5 1mi 5         6 1mi 6        7 1mi 7
   8 1mi 8         9 1mi 9        0xA 1mi ,        0xB 1mi /-/     0xC 1mi ВП      0xD 1mi Cx      0xE 1mi В↑     0xF 1mi FВx
0x10 1mi +      0x11 1mi -       0x12 1mi ×       0x13 1mi ÷      0x14 1mi ↔      0x15 1mi F10ˣ   0x16 1mi Feˣ   0x17 1mi Flg
0x18 1mi Fln    0x19 1mi Farcsin 0x1A 1mi Farccos 0x1B 1mi Farctg 0x1C 1mi Fsin   0x1D 1mi Fcos   0x1E 1mi Ftg
0x20 1mi Fπ     0x21 1mi F√      0x22 1mi Fx²     0x23 1mi F1/x   0x24 1mi Fxʸ    0x25 1mi Fo     0x26 1mi КМ→Г  0x27 1mi K-
0x28 1mi КИПРГ  0x29 1mi К÷      0x2A 1mi КМС→Г
0x30 1mi КГ→МС  0x31 1mi К|x|    0x32 1mi КЗН     0x33 1mi КГ→М   0x34 1mi К[x]   0x35 1mi К{x}   0x36 1mi Кmax  0x37 1mi КAND
0x38 1mi КOR    0x39 1mi КXOR    0x3A 1mi КNOT    0x3B 1mi КСЧ                                   
0x40 1mi П0     0x41 1mi П1      0x42 1mi П2      0x43 1mi П3     0x44 1mi П4     0x45 1mi П5     0x46 1mi П6    0x47 1mi П7
0x48 1mi П8     0x49 1mi П9      0x4A 1mi ПА      0x4B 1mi ПВ     0x4C 1mi ПС     0x4D 1mi ПД     0x4E 1mi ПЕ
0x50 1mi С/П    0x51 &2mi &БП    0x52 1mi В/О     0x53 &2mi &ПП   0x54 1mi КНОП   0x55 1mi КЭКР   0x56 1mi КГРФ  0x57 &2mi &Fx≠0
0x58 &2mi &FL2  0x59 &2mi &Fx≥0  0x5A &2mi &FL3   0x5B &2mi &FL1  0x5C &2mi &Fx<0 0x5D &2mi &FL0  0x5E &2mi &Fx=0
0x60 1mi ИП0    0x61 1mi ИП1     0x62 1mi ИП2     0x63 1mi ИП3    0x64 1mi ИП4    0x65 1mi ИП5    0x66 1mi ИП6   0x67 1mi ИП7
0x68 1mi ИП8    0x69 1mi ИП9     0x6A 1mi ИПА     0x6B 1mi ИПВ    0x6C 1mi ИПС    0x6D 1mi ИПД    0x6E 1mi ИПЕ
0x70 1mi Кx≠00  0x71 1mi Кx≠01   0x72 1mi Кx≠02   0x73 1mi Кx≠03  0x74 1mi Кx≠04  0x75 1mi Кx≠05  0x76 1mi Кx≠06 0x77 1mi Кx≠07
0x78 1mi Кx≠08  0x79 1mi Кx≠09   0x7A 1mi Кx≠0А   0x7B 1mi Кx≠0В  0x7C 1mi Кx≠0С  0x7D 1mi Кx≠0Д  0x7E 1mi Кx≠0Е 0x7F 2mi РКx≠0
0x80 1mi КБП0   0x81 1mi КБП1    0x82 1mi КБП2    0x83 1mi КБП3   0x84 1mi КБП4   0x85 1mi КБП5   0x86 1mi КБП6  0x87 1mi КБП7
0x88 1mi КБП8   0x89 1mi КБП9    0x8A 1mi КБПА    0x8B 1mi КБПВ   0x8C 1mi КБПС   0x8D 1mi КБПД   0x8E 1mi КБПЕ  0x8F 2mi РКБП
0x90 1mi Кx≥00  0x91 1mi Кx≥01   0x92 1mi Кx≥02   0x93 1mi Кx≥03  0x94 1mi Кx≥04  0x95 1mi Кx≥05  0x96 1mi Кx≥06 0x97 1mi Кx≥07
0x98 1mi Кx≥08  0x99 1mi Кx≥09   0x9A 1mi Кx≥0А   0x9B 1mi Кx≥0В  0x9C 1mi Кx≥0С  0x9D 1mi Кx≥0Д  0x9E 1mi Кx≥0Е 0x9F 2mi РКx≥0
0xA0 1mi КПП0   0xA1 1mi КПП1    0xA2 1mi КПП2    0xA3 1mi КПП3   0xA4 1mi КПП4   0xA5 1mi КПП5   0xA6 1mi КПП6  0xA7 1mi КПП7
0xA8 1mi КПП8   0xA9 1mi КПП9    0xAA 1mi КППА    0xAB 1mi КППВ   0xAC 1mi КППС   0xAD 1mi КППД   0xAE 1mi КППЕ  0xAF 2mi РКПП
0xB0 1mi КП0    0xB1 1mi КП1     0xB2 1mi КП2     0xB3 1mi КП3    0xB4 1mi КП4    0xB5 1mi КП5    0xB6 1mi КП6   0xB7 1mi КП7
0xB8 1mi КП8    0xB9 1mi КП9     0xBA 1mi КПА     0xBB 1mi КПВ    0xBC 1mi КПС    0xBD 1mi КПД    0xBE 1mi КПЕ   0xBF 2mi РКП
0xC0 1mi Кx<00  0xC1 1mi Кx<01   0xC2 1mi Кx<02   0xC3 1mi Кx<03  0xC4 1mi Кx<04  0xC5 1mi Кx<05  0xC6 1mi Кx<06 0xC7 1mi Кx<07
0xC8 1mi Кx<08  0xC9 1mi Кx<09   0xCA 1mi Кx<0А   0xCB 1mi Кx<0В  0xCC 1mi Кx<0С  0xCD 1mi Кx<0Д  0xCE 1mi Кx<0Е 0xCF 2mi РКx<0
0xD0 1mi КИП0   0xD1 1mi КИП1    0xD2 1mi КИП2    0xD3 1mi КИП3   0xD4 1mi КИП4   0xD5 1mi КИП5   0xD6 1mi КИП6  0xD7 1mi КИП7
0xD8 1mi КИП8   0xD9 1mi КИП9    0xDA 1mi КИПА    0xDB 1mi КИПВ   0xDC 1mi КИПС   0xDD 1mi КИПД   0xDE 1mi КИПЕ  0xDF 2mi РКИП
0xE0 1mi Кx=00  0xE1 1mi Кx=01   0xE2 1mi Кx=02   0xE3 1mi Кx=03  0xE4 1mi Кx=04  0xE5 1mi Кx=05  0xE6 1mi Кx=06 0xE7 1mi Кx=07
0xE8 1mi Кx=08  0xE9 1mi Кx=09   0xEA 1mi Кx=0А   0xEB 1mi Кx=0В  0xEC 1mi Кx=0С  0xED 1mi Кx=0Д  0xEE 1mi Кx=0Е 0xEF 2mi РКx=0
                0xF1 &3mi &РБП   0xF2 1mi РРВ/О   0xF3 &3mi &РПП                                                 0xF7 &3mi &Px≠0
0xF8 &3mi &PFL2 0xF9 &3mi &Рx≥0  0xFA &3mi &PFL3  0xFB &3mi &PFL1 0xFC &3mi &Px<0 0xFD &3mi &PFL0 0xFE &3mi &Px=0

: next; ( -- )  КБП9 end-code ;
: БП;  ( a -- )  БП end-code ;


\ -------------------------------------
\ Каллисто-2: Байтовые регистры МК-161.
\ -------------------------------------
ONLY FORTH  ALSO VMETA  ALSO VTARGET  DEFINITIONS
b: bufKbd  8 ballot    \ Буфер клавиатуры размером в 8 байт
breg kbdHead           \ Голова буфера клавиатуры: 0..7
breg kbdTail           \ Хвост буфера клавиатуры: 0..7
breg bKey              \ Код нажатой клавиши (или 0)
breg flgKbdCaps        \ Флаг клавиатуры: строчные / заглавные = 0 / 1
breg bKbdCodeFK        \ Предыдущая цифра при FK xx

breg flgCsr            \ Флаг курсора, для мигания
b: bCHW     8 b,       \ Максимальная ширина литеры при выводе на индикатор
b: bCHH     8 b,       \ Высота литеры при выводе на индикатор
b: bBSW     6 b,       \ Ширина средней литеры для BS
b: bSCRLN   1 b,       \ На сколько линий поднять экран при прокрутке
b: bSCRLFIX 0 b,       \ На сколько строк поднять курсор после прокрутки

breg bPocketCsr        \ Позиция курсора в кармане
breg bPocketSize       \ Количество литер в кармане
b: bufPocket 90 ballot \ Карман для ACCEPT
b: bufTIB 90 ballot    \ TIB: буфер входой строки

\ ---------------------------------------
\ Каллисто-2: Десятичные регистры МК-161.
\ ---------------------------------------
\ 9007 19 d!          \ R19=9007: Начинаем в DARK, тёмным по светлому
dreg ptrKbdInt        \ Адрес активного режима клавиатуры

\ -----------------------------------
\ Каллисто-2: Память программ МК-161.
\ -----------------------------------
ONLY FORTH  ALSO VMETA  

start-code
Label &Autoexec

Cx                           \ === Запускаем шитый код ===
5 0 9 0 П2                   \ Инициализация стека возвратов
9 9 8 П3                     \ Инициализация стека данных
&Autoexec &&&&, П6           \ RI := Начало шитого кода - 1

p: NextIni
9 5 0 0                      \ Адрес таблицы свёрток
9042 П                       \ Индексный регистр памяти программ
Fo КБП9                      \ R9 содержит адрес точки NEXT

p: CALL                      \ Обработчик слов высокого уровня
ИП6 В↑ ВП 2 /-/ К[x] КП2     \ Сохраняем старшую половину RI
ВП 2 - КП2 Fo                \ Сохраняем младшую половину RI
ИП7 П6 Fo                    \ RI := PFA - 1 = CFA
                             \ Для быстродействия CALL склеено с NEXT

p: NEXT                      \ Адресный интерпретатор: выполнить следующий токен в шитом коде
КИП6 ВП 2 КИП6 + П7 Fo       \ W := REGW[RI++], считываем CFA в R7
КИП7 9210 П                  \ Находим по CFA адрес обработчика
П8 Fo КБП8                   \ Передаём на него управление
\ 1. Можно сделать если первый КИП6 даёт ноль, то второй КИП6 даёт сразу свёртку -- без CFA
\ такой подход сэкономит сотню-другую байт в памяти данных, но чуть замедлит Каллисто
\ и запутает компилятор
\ 2. Второй точкой входа может быть команда Fo перед NEXT

CALL ps,                     \ Обработчик 000: CALL

\ -------------------
\ Драйвер клавиатуры.
\ -------------------
begin
p: ChPutBell                 \ Издать короткий звуковой сигнал
  4 4 0 В↑ 9                 \ См. BELL -- зависимость от стековой картинки
  9052 П
  РРВ/О                      \ Может завершать прерывание клавиатуры или таймера

p: Timer1Int                 \ Прерывание: таймер 1, автоповтор клавиш
  2 0 9051 П                 \ 200 мс, 5 раз/секунду

p: KeyAdd
  kbdHead ИП                 \ Голова очереди клавиатуры
  kbdTail ИП ПА 1 + 7 КAND   \ Сдвинуть хвост на 1 байт
  - x≠0                      \ Если совпадёт с головой, недовольно пискнуть
until

FВx kbdTail П                \ Сдвинуть хвост
bufKbd pnum, ИПА + ПА        \ RA: куда записать нажатую клавишу
bKey ИП КПА                  \ Записать в очередь клавиатуры код нажатой клавиши
p: Iret
РРВ/О                        \ Возврат из прерывания

p: tblKeyNum                 \ Кнопка в литеры, режим Num
ptext" 0123456789,. "        \ Кнопки 0..9 , /-/ ВП
08 p, 24 p,                  \ Кнопки Cx В↑
ptext" +-"                   \ Кнопки + -
179 p, 176 p, 183 p,         \ Кнопки × ÷ ↔
00 p, 00 p, 00 p,            \ Кнопки F K P
ptext" ():;@![]"
00 p, 27 p, 13 p,            \ Кнопки Р-ГРД-Г ВЫХОД ВВОД
06 p, 02 p, 16 p, 14 p,      \ Выбор ^f ^b ^p ^n

p: tblKeyNumF                \ Кнопка в литеры, режим Num с префиксом F
190 p,                       \ Клавиша ˣ
ptext" '" 34 p,              \ Клавиши ' "
ptext" #$%^&*~"              \ Кнопки 3..9
181 p, ptext" \" 09 p,       \ Клавиши , . пробел
127 p, ptext" |"             \ Кнопки Сx В↑
185 p, 251 p, 189 p,         \ Клавиши π √ ²
ptext" /" 191 p,             \ Кнопки ÷ ↔
00 p, 00 p, 00 p,            \ Кнопки F K P
ptext" <=" 177 p, 178 p,     \ Кнопки ШГ> ШГ< В/О С/П в < = ≥ ≠
ptext" >?{}"                 \ Кнопки ИП П БП ПП
00 p, 200 p, 192 p,          \ Кнопки Р-ГРД-Г ВЫХОД ВВОД
05 p, 01 p, 18 p, 20 p,      \ Выбор ^e ^a ^r ^t

p: tblKeyLat                 \ Кнопка в литеры, режим ЛАТ
ptext" xstunopijkyz " 08 p,  \ Кнопки 0..9 , ВП Сx
ptext" wqlrmv"               \ Кнопки В↑ + - × ÷ ↔
00 p, ptext" d" 00 p,        \ Кнопки F K P
ptext" _abcefgh"             \ Кнопки ШГ> ШГ< В/О С/П ИП П БП ПП
00 p, 27 p, 13 p,            \ Кнопки Р-ГРД-Г ВЫХОД ВВОД
06 p, 02 p, 16 p, 14 p,      \ Выбор ^f ^b ^p ^n

p: tblKeyLatF                \ Кнопка в литеры, режим ЛАТ с префиксом F
ptext" XSTUNOPIJKYZ"         \ Кнопки 0..9 ,
09 p, 127 p,                 \ Кнопки ВП Сx
ptext" WQLRMV"               \ Кнопки В↑ + - × ÷ ↔
00 p, ptext" D" 00 p,        \ Кнопки F K P
ptext" `ABCEFGH"             \ Кнопки ШГ> ШГ< В/О С/П ИП П БП ПП
00 p, 200 p, 192 p,          \ Кнопки Р-ГРД-Г ВЫХОД ВВОД
05 p, 01 p, 18 p, 20 p,      \ Выбор ^e ^a ^r ^t

p: tblKeyCtrl                \ Кнопка в литеры, режим Num с префиксом K
24 p, 19 p, 20 p, 21 p,             \ ^X ^S ^T ^U
14 p, 15 p, 16 p,                   \ ^N ^O ^P
09 p, 10 p, 11 p, 25 p, 26 p,       \ ^I ^J ^K ^Y ^Z
31 p, 28 p,                         \ K ВП и К Cx это ^_ и ^\
23 p, 17 p, 12 p, 18 p, 13 p, 22 p, \ ^W ^Q ^L ^R ^M ^V
255 p, 04 p, 127 p,                 \ K F, K K и K P это 255, ^D и 127
00 p, 01 p, 02 p, 03 p,             \ ^@ ^A ^B ^C
05 p, 06 p, 07 p, 08 p,             \ ^E ^F ^G ^H
30 p, 27 p, 29 p,                   \ K Р-ГРД-Г, К ВЫХОД и К ВВОД это ^^, ^[ и ^]
26 p, 27 p, 24 p, 25 p,             \ Выбор

p: tblKeyRus                 \ Кнопка в литеры, режим РУС
237 p, 232 p, 233 p, 234 p,  \ Клавиши э ш щ ъ
227 p, 228 p, 229 p,         \ Клавиши у ф х
174 p, 175 p, 224 p,         \ Клавиши о п р
238 p, 239 p, 32 p, 08 p,    \ Клавиши ю я пробел BS
236 p, 230 p, 225 p, 231 p,  \ Клавиши ь ц с ч
226 p, 235 p,                \ Клавиши т ы
00 p, 241 p, 00 p,           \ Кнопка F клавиша ё кнопка P
166 p, 167 p, 168 p, 169 p,  \ Клавиши ж з и й
170 p, 171 p, 172 p, 173 p,  \ Клавиши к л м н
00 p, 164 p, 165 p,          \ Кнопка Р-ГРД-Г клавиши д е
162 p, 161 p, 160 p, 163 p,  \ Клавиши в б а г

p: tblKeyRusF                \ Кнопка в литеры, режим РУС с префиксом F
157 p, 152 p, 153 p, 154 p,  \ Клавиши Э Ш Щ Ъ
147 p, 148 p, 149 p,         \ Клавиши У Ф Х
142 p, 143 p, 144 p,         \ Клавиши О П Р
158 p, 159 p, 09 p, 127 p,   \ Клавиши Ю Я ТАБ ЗБ
156 p, 150 p, 145 p, 151 p,  \ Клавиши Ь Ц С Ч
146 p, 155 p,                \ Клавиши Т Ы
00 p, 240 p, 00 p,           \ Кнопка F клавиша Ё кнопка P
134 p, 135 p, 136 p, 137 p,  \ Клавиши Ж З И Й
138 p, 139 p, 140 p, 141 p,  \ Клавиши К Л М Н
00 p, 132 p, 133 p,          \ Кнопка Р-ГРД-Г клавиши Д Е
130 p, 129 p, 128 p, 131 p,  \ Клавиши В Б А Г

Label &KPRus   Label &KPRusF \ Объявляем обработчики нажатий на клавиши
Label &KPLat   Label &KPLatF
Label &KPNum   Label &KPNumF
Label &KPCtrl  Label &KPFK

p: KeyPress                  \ Прерывание: нажатие на кнопку клавиатуры
  9029 ИП  ПА                \ RA := код нажатой кнопки
  ptrKbdInt РКБП             \ Обработать в соответствии со состоянием клавиатуры

p: KP2FK  &KPFK &&&,         \ Нажали FK в состоянии Num
p: KPStateExit
     ptrKbdInt П

p: KeyRelease                \ Прерывание: отпускание кнопки клавиатуры
  Cx bKey П                  \ Сбрасываем код нажатой клавиши
  9151 П                     \ Запретить аппаратное прерывание от таймера 1
  9051 П                     \ Обнуляем таймер 1
  РРВ/О                      \ Возврат из прерывания

p: KP2Num   &KPNum &&&,   KPStateExit БП
p: KP2NumF  &KPNumF &&&,  KPStateExit БП
p: KP2Rus   &KPRus &&&,   KPStateExit БП
p: KP2RusF  &KPRusF &&&,  KPStateExit БП
p: KP2Lat   &KPLat &&&,   KPStateExit БП
p: KP2LatF  &KPLatF &&&,  KPStateExit БП
p: KP2Ctrl  &KPCtrl &&&,  KPStateExit БП

p: KPCaps  1 flgKbdCaps ИП - flgKbdCaps П  KeyRelease БП  \ Нажали Caps в состоянии Num F

&KPCtrl &:                   \ -- Обработать клавишу в состоянии F K xx
  &KPNum &&&, ptrKbdInt П
  Cx tblKeyCtrl pnum, + КИПРГ

p: KPStore
  bKey П                   \ Записать код литеры в bKey
  5 0 9051 П               \ Задержка 500 мс (1/2 с) перед повтором
  3 9151 П                 \ Регистр разрешения аппаратных прерываний: окончание счёта таймера 1: прерывание 3
  KeyAdd БП                \ Добавить клавишу в очередь

Label &KPR1
Label &KPRF1

&KPRus  &:                   \ -- Обработать клавишу в состоянии РУС
  1 2 - &KPR1 &Fx≠0          \ Кнопка 12: Клавиша пробел
  1 - &KPR1 &Fx≠0            \ Кнопка 13: Клавиша BS
  7 - KP2RusF Fx≠0           \ Кнопка 20: Клавиша F
  2 - KP2Num Fx≠0            \ Кнопка 22: Клавиша Num
  9 - KP2Lat Fx≠0            \ Кнопка 31: Клавиша ЛАТ
  flgKbdCaps ИП  &KPRF1 &Px=0
&KPR1 &:
  tblKeyRus pnum,
p: KPFetchTbl
  ИПА + КИПРГ  KPStore Fx=0
  KeyRelease БП

&KPRusF &:                   \ -- Обработать клавишу в состоянии РУС F
  &KPRus pnum, ptrKbdInt П   \ Вернуть клавиатуру в состояние РУС
  Cx 1 2 - &KPRF1 &Px≠0      \ Кнопка 12: Клавиша ТАБ
  1 - &KPRF1 &Px≠0           \ Кнопка 13: Клавиша ЗБ
  9 - KPCaps Fx≠0            \ Кнопка 22: Клавиша Caps
  flgKbdCaps ИП  &KPR1 Fx=0
&KPRF1 &:
  tblKeyRusF pnum,  KPFetchTbl БП

Label &KPL1
Label &KPLF1

&KPLat  &:                   \ -- Обработать клавишу в состоянии ЛАТ
  1 2 - &KPL1 &Fx≠0          \ Кнопка 12: Клавиша пробел
  1 - &KPL1 &Fx≠0            \ Кнопка 13: Клавиша BS
  7 - KP2LatF Fx≠0           \ Кнопка 20: Клавиша F
  2 - KP2Num Fx≠0            \ Кнопка 22: Клавиша Num
  1 - &KPL1 &Fx≠0            \ Кнопка 23: Клавиша _
  8 - &KPL1 &Fx<0            \ Кнопки 31 и выше
  flgKbdCaps ИП  &KPLF1 &Px=0
&KPL1 &:
  tblKeyLat pnum,  KPFetchTbl БП

&KPLatF &:                   \ -- Обработать клавишу в состоянии ЛАТ F
  &KPLat pnum, ptrKbdInt П   \ Вернуть клавиатуру в состояние ЛАТ
  Cx 1 2 - &KPLF1 &Px≠0      \ Кнопка 12: Клавиша пробел
  1 - &KPLF1 &Px≠0           \ Кнопка 13: Клавиша BS
  9 - KPCaps Fx≠0            \ Кнопка 22: Клавиша Caps
  1 - &KPLF1 &Px≠0           \ Кнопка 23: Клавиша `
  8 - KP2Rus Fx≠0            \ Кнопка 31: Клавиша РУС
  &KPLF1 &Px<0               \ Кнопки 32 и выше
  flgKbdCaps ИП  &KPL1 Fx=0
&KPLF1 &:
  tblKeyLatF pnum,  KPFetchTbl БП

&KPNum &:                    \ -- Обработать клавишу в состоянии Num
  tblKeyNum pnum, + КИПРГ  KPStore Fx=0    \ Перевести код кнопки в код литеры
  ИПА 2 0 - KP2NumF Fx≠0     \ Кнопка 20: Клавиша F
  1 - KP2Ctrl Fx≠0           \ Кнопка 21: Клавиша K
  1 0 - KP2Lat Fx≠0          \ Кнопка 31: Клавиша ЛАТ
  KeyRelease БП

&KPNumF &:                   \ -- Обработать клавишу в состоянии Num F
  &KPNum pnum, ptrKbdInt П   \ Вернуть клавиатуру в состояние Num
  Cx tblKeyNumF pnum, + КИПРГ  KPStore Fx=0   \ Перевести код кнопки в код литеры
  ИПА 2 1 - KP2FK Fx≠0       \ Кнопка 21: Клавиша K
  1 - KPCaps Fx≠0            \ Кнопка 22: Клавиша Caps
  9 - KP2Rus Fx≠0            \ Кнопка 31: Клавиша РУС
  KeyRelease БП

p: KTH1  ИПА
p: KTH2  В/О
p: KTH4  1 3 + В/О
p: KeyToHex
  1 0 - KTH1 Fx≥0
  1 1 - KTH4 Fx≠0   \ Кнопка 21 это D = 13
  3 - KTH2 Fx≥0     \ Кнопки 22 и 23 не используются
  3 - KTH4 Fx≥0     \ Кнопки 24, 25 и 26 это A, B и C
  1 ↔ - KTH2 Fx≥0
  1 5 ↔ - В/О       \ Кнопки 27 и 28 это E и F

p: KPFK2
  &KPNum pnum, ptrKbdInt П
  ИПА KeyToHex ПП  KeyRelease Fx≥0
  bKbdCodeFK ИП  1 6 × + KPStore БП

&KPFK   &:                   \ -- Обработать клавишу в состоянии F K xx
  KeyToHex ПП  bKbdCodeFK П  KP2Num Fx≥0
  KPFK2 pnum,  KPStateExit БП

\ CHGET -- Ввод литеры в RX с терминала

p: CsrOff
  Cx 9150 П                  \ Запретить аппаратное прерывание от таймера 0
  9050 П                     \ Обнуляем таймер 0
  flgCsr ИП
  x≠0 if                     \ Стираем курсор, если он выведен
  Cx
p: CsrBlinkM  flgCsr П
  9001 ИП ↔                  \ Сохраняем атрибут вывода в RX1
  3 9001 П                   \ Атрибут 3: XOR
  bCHH ИП 1 9013 П           \ Выводим курсор
  КГРФ                       \ Экран на индикатор
  FВx 9001 П                 \ Восстанавливаем атрибут вывода
  then
  В/О

p: CsrOn
  1 CsrBlinkM ПП             \ 1: курсор включён
  5 0 9050 П                 \ 500 мс, 2 раза/секунду
  4 9150 П                   \ Регистр разрешения аппаратных прерываний: окончание счёта таймера 0: прерывание 4
  В/О

p: Timer0Int                 \ Прерывание: таймер 0, мигание курсора
  5 0 9050 П                 \ 500 мс, 2 раза/секунду
  1 flgCsr ИП -  CsrBlinkM ПП
  РРВ/О                      \ Возврат из прерывания

p: ChGet                     \ Ждём нажатия клавиши и берём её код из очереди клавиатуры
CsrOn ПП
begin
  kbdTail ИП kbdHead ИП - x≠0
until
bufKbd pnum, FВx + ПА КИПА ПА
FВx 1 + 7 КAND kbdHead П
CsrOff ПП
ИПА В/О

\ CHPUT -- Вывод литеры из RX на терминал
\ Портит RA и R9042 (важно, потребует восстановления), не делает КГРФ

Label &tblCHPUT
p: ChPutScroll bSCRLN ИП 19 РКП  9000 ИП  ↔ bSCRLFIX ИП - ↔ 9000 П  В/О
p: ChPutCR     9000 ИП  Cx  9000 П  В/О
p: ChPutCLS    8 19 РКП \ Склейка
p: ChPutHome   Cx В↑ 9000 П  В/О
p: ChPutLF     9000 ИП ↔ bCHH ИП + 6 5 FВx - - ChPutScroll Fx<0 FВx + ↔ 9000 П В/О
p: ChPutTAB    9000 ИП  3 2 +  2 2 4 КAND  9000 П  В/О
p: ChPutBS     9000 ИП  bBSW ИП -  0 Кmax ↔ Fo  9000 П  В/О

p: ChPut
  ПА  &tblCHPUT &&&&, 9042 П  ↔
  9212 П x≥0 if ПА КБПА then   \ Обработка управляющих символов
  9000 ИП 1 2 9 bCHW ИП - -    \ Грубая проверка, что литера уместится
  x≥0 if Cx 9000 П             \ CR, по достижении конца строки
         ChPutLF ПП            \ LF
  then
  ИПА 9020 П В/О               \ Вывести литеру

&tblCHPUT &:                   \ Управляющие символы
 07 p, ChPutBell pw,           \  7 Bell KG  Звонок терминала
 08 p, ChPutBS pw,             \  8 BS   KH  Пробел назад
 09 p, ChPutTAB pw,            \  9 TAB  KI  Горизонтальная табуляция
 10 p, ChPutLF pw,             \ 10 LF   KJ  Перевод строки
 11 p, ChPutHome pw,           \ 11 Home KK  Курсор в левый верхний угол
 12 p, ChPutCLS pw,            \ 12 CLS  KL  Очистка экрана
 13 p, ChPutCR pw,             \ 13 CR   KM  Возврат каретки
 255 p,

\ Вспомогательные подпрограммы Каллисто-2
begin
  FВx - В/О
p: Sign32                      \ Преобразование 32-битного беззнакового в 32-битное со знаком.
  3 2 7 6 8  В↑  6 5 5 3 6  ×  -
  x<0
until
FВx + В/О

end-code

\ Основной код инициализации Каллисто-2
4880 1412 #code INIT           \ === Примитив INIT ===
  1 9148 П                     \ Сброс регистров прерываний
  KeyPress pnum, В↑ 1 2 7      \ Вектор KeyPress, приоритет 127
  9131 П                       \  → регистр настройки прерывания 1
  1 9153 П                     \ Регистр разрешения аппаратных прерываний: нажатие на кнопку клавиатуры: прерывание 1
  KeyRelease pnum, В↑ 1 2 8    \ Вектор KeyRelease, приоритет 128
  9132 П                       \  → регистр настройки прерывания 2
  2 9157 П                     \ Регистр разрешения аппаратных прерываний: отпускание кнопки клавиатуры: прерывание 2
  Timer1Int pnum, В↑ 1 2 6     \ Вектор Timer1Int, приоритет 126 (автоповтор)
  9133 П                       \  → регистр настройки прерывания 3
  Timer0Int pnum, В↑ 1 2 5     \ Вектор Timer0Int, приоритет 125 (мигание курсора)
  9134 П                       \  → регистр настройки прерывания 4
  1 9130 П                     \ Общее разрешение прерываний: прерывания разрешены
  2 9010 П                     \ Инициализация графического экрана
next;                          \ NEXT

4420 6755 #code BYE  ( -- )    С/П  next; \ Вообще-то после перезапуска ещё надо клавиатуру инициализировать INIT

4619 3284 #code EXIT  ( -- )              \ === Примитив EXIT ===
  ИП2 П8 П5  2 + П2 Fo         \ RP += 2
  КИП5 ВП 2 КИП8 + П6 Fo       \ RPOP RI
next;                          \ NEXT

start-code
  begin
    FВx + КБП9
end-code
2689 3609 #code (LITERAL)  ( -- x )
    ↔ КП3 ↔
    КИП6 2 5 6 × КИП6 +
    3 2 7 6 8 -  x≥0
  until
  FВx -
next;

code litb  ( -- c )    ↔ КП3 ↔ КИП6 next;
0537 7636 #code ИПРГ  ( u -- c )  КИПРГ next;

\ Операции со стеком параметров
0000 3297 #code ↔     ( y x -- x y )     ↔ next;
\ code 2swap  ИП3 П4 П5  КИП4 ИПС КП5 ↔ ПС  КИП4 03 РКИП КП5 ↔ 03 РКП  next;
5281 9524 #code OVER  ( y x -- y x y )   ↔ КП3  next;
4211 9684 #code ?DUP ( x -- 0 | x x )    Кx≠09  end-code   \ Здесь склейка: нет NEXT
0000 3777 #code ↑    ( x -- x x )        ↔ КП3  p: jdup2  ↔ В↑  next;
5342 2948 #code PICK ( xu .. x1 x0 u -- xu .. x1 x0 xu )  Кx≥09 jdup2 Fx≠0  ИП3 + ПС ↔ КИПС  next;
3349 3508 #code 2DUP ( y x -- y x y x )  ↔ КП3  ↔ КП3  next;
5477 7987 #code ROT  ( z y x -- y x z )  ↔ ↔ 03 РКИП  FВx 03 РКП  Fo next;
4550 1956 #code DROP ( x -- )            p: jdrop  ↔  p: jdrop2  03 РКИП ↔  1 ИП3 + П3 Fo next;
3349 3509 #code 2DROP ( y x -- )         p: j2drop  2  p: ndrop  ИП3 + П3 П0  КИП0 КИП0  next;

\ Условные операторы
0022 3138 #code 0=  ( x -- f )  x≠0
  if  p: STF  Cx  КБП9
0022 3170 #code 0>  ( x -- f )  /-/ STF Fx<0
  then
  p: STT  Cx 1 /-/  next;

0022 2466 #code 0<  ( x -- f )  STT Fx≥0 Cx next;

0000 1953 #code =   ( y x -- f )  - x≠0
  if  p: STFP  Cx  jdrop2 БП
0000 1921 #code <   ( y x -- f )  - STFP Fx<0
  then
  p: STTP  1 /-/  jdrop2 БП;

0000 1985 #code >   ( y x -- f )  - STTP Fx≥0  STFP БП;

\ Некоторые константы можно позже реализовать через CONSTANT
4675 0373 #code FALSE   ( -- 0 )     ↔ КП3 Cx          next;
0000 1569 #code 1       ( -- 1 )     ↔ КП3 Cx ВП       next;
0000 1601 #code 2       ( -- 2 )     ↔ КП3 Cx 2        next;
5611 8436 #code TRUE    ( -- -1 )    ↔ КП3 Cx 1 /-/    next;
0000 3361 #code π       ( -- π )     ↔ КП3 ↔ Fπ        next;
0030 6562 #code BL      ( -- 32 )    ↔ КП3 Cx 3 2      next;
4467 6995 #code C/L     ( -- 64 )    ↔ КП3 Cx 6 4      next;
4401 3253 #code B/BUF   ( -- 3072 )  ↔ КП3 Cx 3 0 7 2  next;
5738 5927 #code VERSION ( -- n )     ↔ КП3 Cx 2 1 /-/  next;
5607 6867 #code TIB     ( -- a )     bufTIB pnum,      next;

\ Арифметика
0022 7170 #code 1+  ( x -- x+1 )    1 +   next;
0022 7234 #code 1-  ( x -- x-1 )    1 -   next;
0023 1778 #code 2+  ( x -- x+2 )    2 +   next;
0023 1842 #code 2-  ( x -- x-2 )    2 -   next;
0023 3570 #code 2×  ( x -- x×2 )    В↑ +  next;
0023 1906 #code 2/  ( x -- x/2 )    2 ÷ К[x]  next;
0000 1377 #code +   ( y x -- y+x )  +  jdrop2 БП;
0000 1441 #code -   ( y x -- y-x )  -  jdrop2 БП;
0000 3169 #code ×   ( y x -- y×x )  ×  jdrop2 БП;
0000 3073 #code ÷   ( y x -- y÷x )  ÷  jdrop2 БП;
0000 1505 #code /   ( y x -- y/x )  ÷ К[x]  jdrop2 БП;
5145 9715 #code MOD ( y x -- x1 )   ↔ В↑ FВx ÷ FВx ↔ К[x] × -  jdrop2 БП;
\ code /mod  03 РКИП В↑ ИПС ÷ FВx ↔ К[x] ПС × - 03 РКП  next;
\ code ×/  ИП3 П4  03 РКИП КИП4 × ИПС ÷ К[x] ПС  2 ndrop БП;
\ code ×/mod  1 ИП3 + П8  03 РКИП КИП8 × В↑ ИПС ÷ FВx ↔ К[x] ПС × - КП8  ИП8 П3  next;

5212 3779 #code NOT  ( n -- i )     К[x]  Sign32 ПП  1 + /-/  next;
5139 5843 #code MAX  ( y x -- x1 )  Кmax    jdrop2 БП;
5143 2387 #code MIN  ( y x -- x1 )  Кmax    jdrop  БП;
0041 3634 #code Yˣ   ( p x -- x1 )  ↔ Fxʸ   jdrop2 БП;

\ Примитивы МК-161
0000 4161 #code √    ( p -- x )    F√    next;
0035 2706 #code LN   ( p -- x )    Fln   next;
0035 2482 #code LG   ( p -- x )    Flg   next;
0040 8994 #code X²   ( x -- p )    Fx²   next;
0032 1474 #code Eˣ   ( x -- p )    Feˣ   next;
3273 8755 #code 10ˣ  ( x -- p )    F10ˣ  next;
9330 7267 #code |X|  ( x -- p )    К|x|  next;
0007 9074 #code СЧ   ( -- p )      ↔ КП3 ↔ КСЧ  next;
0003 2674 #code ЗН   ( x -- n )    КЗН   next;
3139 5811 #code /-/  ( x -- x1 )   /-/   next;
3273 3443 #code 1/X  ( x -- x1 )   F1/x  next;
6079 1715 #code [X]  ( x -- x1 )   К[x]  next;
9264 3747 #code {X}  ( x -- x1 )   К{x}  next;

\ Тригонометрия МК-161
5471 3447 #code RADIANS  ( -- )       1  p: SetAng  9045 П  Fo  next;
4544 2151 #code DEGREES  ( -- )       0  SetAng БП;  \ Нельзя Cx, т.к. верхушки стеков совпадают
4351 1238 #code ARCSIN   ( x -- x1 )  Farcsin  next;
4351 1398 #code ARCCOS   ( x -- x1 )  Farccos  next;
4351 1013 #code ARCTG    ( x -- x1 )  Farctg   next;
5541 3699 #code SIN      ( x -- x1 )  Fsin  next;
4482 4675 #code COS      ( x -- x1 )  Fcos  next;
0038 9346 #code TG       ( x -- x1 )  Ftg   next;

\ Клавиатура
5008 7203 #code KEY  ( -- c )  П8 ↔ КП3  ChGet ПП  ИП8 ↔  next;

\ Дисплей: вывод текста на индикатор
4614 2596 #code EMIT ( c -- )  ↔ П8 ↔ ChPut ПП
  9 5 0 0                      \ Адрес таблицы свёрток
  9042 П                       \ Индексный регистр памяти программ
  ИП8 jdrop2 БП;

0000 3585 #code CR ( -- )                 \ Должно быть ↵, код 192
  П8 ↔ П7  ChPutCR ПП  ChPutLF ПП  ИП7 ИП8 next;

4411 4820 #code BELL ( -- )   ChPutBell ПП Fo Fo next;            \ Сильно зависит от стековой картинки ChPutBell
5076 9029 #code LIGHT ( -- )  4 9001 П Fo  9 0 0 8 19 П Fo  next; \ Светлым по тёмному, тёмная прокрутка
4542 3460 #code DARK  ( -- )  0 9001 П Fo  9 0 0 7 19 П Fo  next; \ Тёмным по светлому, светлая прокрутка
0206 5027 #code ГРФ ( -- )    КГРФ next;
0000 3841 #code PAGE ( -- )   П8 ↔ П7  ChPutCLS ПП  ИП7 ИП8 next; \ Должно быть ↖, код 200

5615 0692 #code TYPE ( a u -- )           \ Печать строки из байтовой памяти
  П0 /-/ j2drop Fx<0           \ Выход, если u<=0
  Cx ВП - П5  begin  КИП5 ChPut ПП  fl0
p: TypeR
  КГРФ
  9 5 0 0                      \ Адрес таблицы свёрток
  9042 П                       \ Индексный регистр памяти программ
  j2drop БП;

5347 3445 #code PTYPE  ( a u -- )         \ Печать строки из памяти программ
  П0 /-/ j2drop Fx<0           \ Выход, если u<=0
  ↔ П5  begin  ИП5 КИПРГ ChPut ПП КИП5  fl0  TypeR БП;

Label &tblAccept
\ R0           Временный регистр обработчика
\ R1           Временный регистр обработчика
\ R2           Временный регистр
\ R3 SP        Указатель десятичного стека: растёт вниз от 998
\ R4           Временный регистр
\ R5           Временный регистр
\ R6 RI, IP    Указатель инструкций, указывает на токен в шитом коде
\ R7           (максимальный) размер буфера
\ R8 JP        временный регистр обработчика
\ R9 =NEXT     Адрес точки NEXT
\ RA           Литера, введённая с клавиатуры (Затирается ChGet)
\ RB           Адрес буфера
\ RC           сохраняем указатель целочисленного стека возвратов R2=RP, растёт вниз от 5090
\ RД           (Общее) количество введённых литер
\ RE           Число символов слева от курсора

start-code
p: AccCTL \ Очистить строку справа от курсора до конца индикатора
  tails 9000 ИП  1 2 8 ↔ -  bCHH ИП ↔  9013 П then
p: AccAtrErase
  4 9001 ИП  КXOR  9001 П           \ Атрибут 4 для DARK, 0 для LIGHT
p: AccR
  В/О \ Может потребоваться смена атрибута

p: AccCut \ Очистить строку до конца строки ввода, перемещает курсор и затирает R0, R8
  ИПЕ ИПД - AccR Fx<0  AccCTL ПП               \ Убедиться, что удаляемый хвост есть
  9000 ИП ↔ П8 bCHH ИП + 0 9000 П              \ R8: строка, с которой удаляем хвост
  AccAtrErase ПП                               \ Атрибут на очистку текста
  ИПС ИПД В↑ + - П0 КИП0 КИП0 ИП8 - bCHH ИП +  \ Сколько строк очистить
  1 2 8 9013 П  AccAtrErase БП                 \ Очистить от текста и восстановить атрибут

\ AccChPut (RA: литера, R1: позиция в буфере координат) затирает R0, R8
p: AccChPut                         \ Вывести на экран литеру из RA
  9000 ИП  1 2 9 bCHW ИП - -        \ Грубая проверка на правую границу индикатора
  x≥0 if                            \ Если курсор у конца, то
    9000 ИП П0 ↔ П8                 \ Сохранить координаты курсора
    ИПА 9020 П  9000 ИП 1 2 7 -     \ Вывести символ и проверить, не упёрлись ли в конец индикатора
    x<0 if                          \ Если места хватило, то
      ИП0 КП1 ИП8 КП1               \ Записать в буфер координаты выведенной литеры
p: AccCPR
      ИП1 П0  9000 ИП КП0 ↔ КП0     \ Записать в буфер предварительные координаты следующей литеры
      В/О
    then \ x<0 (литера таки не влезла)
    ИП8 ИП0 9000 П                  \ Восстановить позицию курсора
    AccCTL ПП                       \ Очистить до конца строки
    9000 ИП ↔ bCHH ИП + 6 5 FВx - - \ Грубая проверка на последнюю строку
    x<0 if                          \ Если ещё место есть внизу,
      FВx + 0  9000 П               \ Курсор в начало следующей строки
    else
      ChPutCR ПП  ChPutScroll ПП    \ Прокрутка экрана
      ИПД x≠0 if П0
        8 bSCRLN ИП × В↑ В↑         \ Стек -- на сколько линий корректируем y при прокрутке
        ИПС                         \ Начало буфера координат, растёт вниз
        begin                       \ Обновляем y в буфере координат
          2 - П8 Fo  КИП8 + КП8 Fo  ИП8
        fl0
      then \ RД≠0
    then   \ x<0 (без прокрутки)
  then     \ x≥0 (без перевода строки)
  9000 ИП КП1 ↔ КП1 ↔               \ Записать в буфер координаты текущей литеры
  ИПА 9020 П                        \ Вывести литеру на экран
  AccCPR БП                         \ Записать в буфер новые координаты
end-code

2684 2920 #code (ACCEPT)  ( a c1 c2 c3 -- c4 c5 )  \ a, c1: буфер; c2, c4 длина строки; c3, c5 позиция курсора
\ Инициализация регистров
  ПЕ                         \ RE:=c3 -- число литер слева от курсора
  ↔ ПД                       \ RД:=c2 -- число литер в заготовке
  03 РКИП П7                 \ R7:=c1 -- размер буфера строки
  2 ИП3 П5 + П3              \ 2DROP
  КИП5 ПВ                    \ RB:=a  -- адрес буфера строки
  &tblAccept &&&&, 9042 П    \ R9042 -- таблица переходов ACCEPT (Индексный регистр памяти программ)
  ИП2 ПС  9000 ИП КП2 ↔ КП2  \ RC:=R2 -- сохраняем RP, начало буфера координат

\ Отобразить заготовку строки и создать буфер координат
  ИПД x≠0 if П2              \ Цикл по всем литерам в строке ввода
    ИПВ 1 - П4               \ R4 -- указатель на очередной символ в буфере строки
    ИПС П1                   \ R1 -- начало буфера координат
p: AccRefresh
    begin
      КИП4 ПА AccChPut ПП    \ Отобразить очередную литеру
    fl2
p: AccLocate
    ИПС  ИПЕ В↑ +  -  П1     \ Находим позицию курсора в буфере координат
    КИП1 КИП1 ↔ 9000 П       \ Позиционируем курсор
  then \ RД≠0

\ Основной цикл (ACCEPT)
p: Acc0
  ChGet ПП  ПА                \ RA: литера, введённая с клавиатуры
  9212 П x≥0 if ПА КБПА then  \ Клавиша редактирования?  Обработать.
  ИПД ИП7 -  Acc0 Fx<0        \ Введены все литеры?
  ИПС  ИПЕ В↑ +  -  П1        \ R1: Текущий адрес в буфере координат
  1 ИПД + ПД                  \ В буфере строки на 1 литеру больше
  1 ИПЕ + ПЕ                  \ Курсор вправо
  - x=0 if                    \ Мы в конце буфера? Возможно, проверка излишня
    ИПВ ИПЕ + П2  ИПА КП2     \ Запишем литеру в строку
    AccChPut ПП               \ Вывести RA на индикатор
    Acc0 БП
  then                        \ Не в конце буфера, надо двигать хвост
  1 + П2                      \ Количество выводимых литер
  ИПВ ИПЕ + 2 - П5            \ R5: указатель на курсор в буфере
  begin
    КИП5 П4                   \ R4: замещаемая литера из буфера
    ИПА 05 РКП  AccChPut ПП   \ Записать литеру в буфер и вывести на экран
    ИП4 ПА                    \ Проделать тоже самое с литерой из буфера
  fl2
  AccLocate БП                \ Курсор в нужную позицию и опять ждём клавиши

p: AccEnter                   \ Клавиша ВВОД
  ИПС  П2                     \ Восстанавливаем R2=RP
  ИПД В↑ +  -  П1             \ Адрес последней позиции в буфере координат
  КИП1 КИП1 ↔ 9000 П          \ Курсор в конец введённой строки
  ИПД ИПЕ  NextIni БП         \ Y, X: Размер введённой строки и позиция курсора

p: AccBOL     Cx 0x5E p,      \ Пропустить один шаг (Fx=0)
p: AccEOL     ИПД
p: AccMLoc    ПЕ  AccLocate БП

p: AccUp      ИПЕ 8 0xF7 p,   \ Пропустить два шага (Px≠0)
p: AccLeft    ИПЕ 1 -  0 Кmax  AccMLoc БП

p: AccDown    8 0x57 p,       \ Пропустить один шаг (Fx≠0)
p: AccRight   1 ИПЕ +  ИПД  Кmax  ↔ AccMLoc БП

p: AccBS                                    \ Клавиша BS
  ИПЕ Acc0 Fx≠0 1 - ПЕ                      \ Если слева символов нет, ничего не делать
  ИПС ИПЕ В↑ +  - П1 П0                     \ R0, R1: новая позиция в буфере координат
  КИП0 КИП0 ↔ 9000 П  AccCut ПП             \ Очистить хвост на экране
  ИПД 1 - ПД                                \ Уменьшаем на 1 количество литер в строке
  ИПЕ - AccLocate Fx≠0 П0 П2                \ Если хвоста нет, всё уже сделано
  ИПВ ИПЕ + П4 1 - П5 П8                    \ R4: откуда; R5, R8: куда
  begin КИП4 КП5 fl0                        \ Сдвинуть хвост
  ИП8 П4  ИП1 П0                            \ R4: откуда выводить символы, R0: позиция в буфере координат
  КИП0 КИП0 ↔ 9000 П  AccRefresh БП         \ Вывести новый хвост на экран

p: AccIns                                   \ Вставить строку из кармана
  ИП7  bPocketSize ИП  ИПД +  -  Acc0 Fx≥0  \ Если слишком много символов в кармане, ничего не делать
  ИПВ ИПД + П2  bPocketSize ИП + П1         \ R2: откуда; R1: куда
  FВx  ИПД ИПЕ - П0  + П8                   \ R0: размер хвоста; R8: размер хвоста + размер кармана
  ИП0 x≠0 if  begin КИП2 КП1 fl0  then      \ Перенести хвост
  ИПД  bPocketSize ИП  П0 + ПД              \ R0: размер кармана; RД: новый размер входной строки
  FВx  bufPocket pnum, + П2                 \ R2: адрес конца кармана + 1
  begin КИП2 КП1 fl0                        \ Вставить карман
  ИП8 П2  ИПВ ИПЕ + 1 - П4                  \ R2: число литер справа от курсора; R4: откуда эти литеры выводить
  ИПС ИПЕ В↑ +  - П1                        \ R1: позиция в буфере координат
  begin  КИП4 ПА AccChPut ПП  fl2           \ Вывести литеры на экран
  ИПЕ bPocketCsr ИП + AccMLoc БП            \ Курсор вправо, где он был в кармане

p: AccSave                                           \ Убрать строку в карман
  ИПЕ bPocketCsr П  ИПД bPocketSize П П0  Acc0 Fx≠0  \ R0: Длина сохраняемой строки, если ноль выйти
  bufPocket 01 [-] pnum, П5  ИПВ 1 - П4              \ R4: откуда; R5: куда
  begin КИП4 КП5 fl0                                 \ Скопировать строку в карман
  0 ПЕ  ИПС П0  КИП0 КИП0 ↔ 9000 П                   \ В начало строки, склейка
p: AccFCx                                            \ Обрубить строку в позиции курсора
  AccCut ПП  ИПЕ ПД  AccLocate БП                    \ Стереть хвост строки
end-code

&tblAccept &:           \ Клавиши редактирования строки
 01 p, AccBOL pw,       \   1  F←    Курсор в начало строки
 02 p, AccLeft pw,      \   2  ←     Курсор влево
 05 p, AccEOL pw,       \   5  F→    Курсор в конец строки
 06 p, AccRight pw,     \   6  →     Курсор направо
 08 p, AccBS pw,        \   8  Cx    BS, удалить символ слева от курсора
 13 p, AccEnter pw,     \  13  Ввод  Завершить редактирование строки
 14 p, AccDown pw,      \  14  ↓     Курсор на 8 символов вправо
 16 p, AccUp pw,        \  16  ↑     Курсор на 8 символов влево
 18 p, AccIns pw,       \  18  F↑    Вставить строку и позицию курсора
 20 p, AccSave pw,      \  20  F↓    Сохранить строку и позицию курсора и сбросить строку
127 p, AccFCx pw,       \ 127  F Cx  Удалить символы справа от курсора
255 p,

:: ACCEPT  ( a c1 -- c2 ) FALSE FALSE (ACCEPT) DROP EXIT ;;

\ Хэширование имени

\     0 1 2 3 4 5 6 7 8 9 a b c d e f
\ 00  * - - - - - - - - - - - - - - -  =>  xx  85 - - - - - - - - - - - - - - -
\ 10  * - - * - - - - ↑ ↓ → ← - - - *  =>  xx  83 - - 84 - - - - 76 77 75 74 - - - 20
\ 20  * ! " # $ % & ' ( ) * + , - . /  => -00  20..2f
\ 30  0 1 2 3 4 5 6 7 8 9 : ; < = > ?  => -00  30..3f
\ 40  @ A B C D E F G H I J K L M N O  => -00  40..4f
\ 50  P Q R S T U V W X Y Z [ \ ] ^ _  => -00  50..5f
\ 60  ` a b c d e f g h i j k l m n o  => -20  80 41..4f
\ 70  p q r s t u v w x y z { | } ~ *  => -20  50..5a  8b 8c 8d 8e 8f
\ 80  А Б В Г Д Е Ж З И Й К Л М Н О П  => -80  00..0f
\ 90  Р С Т У Ф Х Ц Ч Ш Щ Ъ Ы Ь Э Ю Я  => -80  10..1f
\ a0  а б в г д е ж з и й к л м н о п  => -a0  00..0f
\ b0  * * * * * * - * - * - - - * * *  => -50  60..6f
\ c0  * * * * ← → ↑ ↓ * * * - - - - -  => -50  70..7f
\ d0  - - - - - - - - - - - - - - - -  =>  xx  - - - - - - - - - - - - - - - -
\ e0  р с т у ф х ц ч ш щ ъ ы ь э ю я  => -d0  10..1f
\ f0  Ё ё - - - - - - - - - √ - - - -  =>  xx  81 81 - - - - - - - - - 82 - - - -
\
p: tblHash              \ Особые случаи
    00 p, 00 p, 0x85 p, \ 00 * => 85
  0x10 p, 00 p, 0x83 p, \ 10 * => 83
  0x13 p, 00 p, 0x84 p, \ 13 * => 84
  0x18 p, 00 p, 0x76 p, \ 18 ↑  (c6) => 76
  0x19 p, 00 p, 0x77 p, \ 19 ↓  (c7) => 77
  0x1A p, 00 p, 0x75 p, \ 1a →  (c5) => 75
  0x1B p, 00 p, 0x74 p, \ 1b ←  (c4) => 74
  0x1F p, 00 p, 0x20 p, \ 1f * => 20
  0x60 p, 00 p, 0x80 p, \ 60 ' => 80
  0x7B p, 00 p, 0x8B p, \ 7b { => 8b
  0x7C p, 00 p, 0x8C p, \ 7c | => 8c
  0x7D p, 00 p, 0x8D p, \ 7d } => 8d
  0x7E p, 00 p, 0x8E p, \ 7e ~ => 8e
  0x7F p, 00 p, 0x8F p, \ 7f * => 8f
  0xF0 p, 00 p, 0x81 p, \ f0 Ё => 81
  0xF1 p, 00 p, 0x81 p, \ f1 ё => 81
  0xFB p, 00 p, 0x82 p, \ fb √ => 82
   255 p,

\  00   10   20   30   40   50   60   70   80   90   a0   b0   c0   d0   e0   f0
\  xx   xx  -00  -00  -00  -00  -20  -20  -80  -80  -a0  -50  -50   xx  -d0   xx
\  d1   d1   d0   d0   d0   d0   b0   b0   50   50   30   80   80   d1   00   d1
\
\ -d0  00h   0
\ -a0  30h  48
\ -80  50h  80
\ -50  80h 128
\ -20  b0h 176
\ -00  d0h 208
\  xx  d1h 209
p: tblHash2                      \ Преобразование групп символов
  0xD1 p, 0xD1 p, 0xD0 p, 0xD0 p, 0xD0 p, 0xD0 p, 0xB0 p, 0xB0 p,
  0x50 p, 0x50 p, 0x30 p, 0x80 p, 0x80 p, 0xD1 p,   00 p, 0xD1 p,

start-code
p: CHash ( c -- n )              \ Свёртывает литеру c в код от 00 до 8f, -1 при неверной литере
  В↑ В↑  9212 П
  x<0 if                         \ Если литера не найдена в таблице хэшей
    Cx 1 6 ÷ К[x]  tblHash2 pnum, +  КИПРГ  2 0 8 ↔ -
    x≥0 if - then
  then
  В/О
end-code

Label &Hash1
Label &Hash2
4807 7572 #code HASH ( a n -- d )           \ Возвращает хэш d строки a длиной n
  П0 jdrop2 Fx≠0                 \ R0:=n; Если n=0, вернуть хэш 0
  Cx ПС П5 ВП - П4               \ Указатель R4:=a-1; Хэш RC:=0; Счётчик литер R5:=0
  tblHash pnum, 9042 П           \ Таблица особых случаев для CHASH
p: H1L
  КИП4 CHash ПП                  \ Считать кандидат на литеру1
  x<0 if                         \ Засчитать литеру?
    H1L FL0                      \ Взять следующую
    ИПС                          \ Литер нет, хэш 0
p: HRX                           \ Найденый хэш в RX
    9 5 0 0                      \ Адрес таблицы свёрток
    9042 П                       \ Индексный регистр памяти программ
    jdrop БП                     \ Выйти, вернув хэш в RY
  then
  ПС КИП5                        \ RC := л1; R5 := 1
  &Hash1 &РБП

p: H2L
  КИП4 CHash ПП                  \ Считать кандидат на литеру2
  x<0 if                         \ Засчитать литеру?

&Hash1 &:
    H2L FL0                      \ Взять следующую
p: HashOk
    ИПС
p: HashOk2
    3 2 ×  ИП5 +  HRX БП         \ хэш := литеры×32+N
  then
  ИПС 1 4 4 × + ПС               \ RC := л1×144+л2
  КИП5                           \ R5 := 2
  &Hash2 &РБП

p: H3L
  КИП4 CHash ПП                  \ Считать ещё одну литеру
  x≥0 if                         \ Засчитать литеру?
    ПВ                           \ RB := кандидат на л3
    КИП5                         \ R5++, увеличить счётчик символов
    H3L FL0                      \ Продолжим считать литеры и искать последнюю
p: HF
    ИПВ ИПС 1 4 4 × +            \ RX := (л1×144+л2)×144+л3
    HashOk2 БП                   \ Вернуть хэш
  then

&Hash2 &:
  H3L FL0                        \ Взять следующую литеру
  ИП5 2 -  HashOk Fx≠0           \ Если литер было две, хэш (л1×144+л2)×32+2
  HF                             \ Если больше -- учесть л3 в RB
БП;

\ start-code
\ p: EmitHex ( c -- )
\   1 0 - x<0 if 5 8 + ChPut БП then
\   6 5 + ChPut БП
\ end-code

\ code HexOut ( c -- )
\   ↔ ПВ ↔
\   1 6 ÷ ПС К[x] EmitHex ПП
\   ИПС К{x} 1 6 × EmitHex ПП
\   3 2  ChPut ПП КГРФ
\   9 5 0 0                      \ Адрес таблицы свёрток
\   9042 П                       \ Индексный регистр памяти программ
\   ИПВ jdrop2
\ БП;

0000 1473 #code . ( d -- )  9023 П  jdrop БП;

:: testAAA                       \ Определение высокого уровня, проверка работоспособности
\    test
    VERSION . litb 62 b, EMIT    \ -nn>
    TIB  litb 90 b,
    ACCEPT CR
    TIB ↔
    2DUP TYPE CR                 \ Тестовая печать
    HASH .
    KEY DROP
    EXIT
;;

p: strForth
12 b,                            \ CLS
ptext" Draft"

\ ------------------------------------------------------
\ Каллисто-2: Инициализация десятичных регистров МК-161.
\ ------------------------------------------------------

NEXT 09 d!                       \ Можно ещё выделить регистры для "других" NEXT
bc [1-] &Autoexec &!             \ Адрес автозапуска шитого кода (ссылка вперёд)
&KPNum ptrKbdInt d!              \ Клавиатура по умолчанию в режиме Num

\ -------------------------------------
\ Каллисто-2: Байтовые регистры МК-161.
\ -------------------------------------

\ === Автозапускающийся шитый код ===
INIT                             \ Начальная инициализация
DARK                             \ Тёмным по светлому
(LITERAL) strForth bw,           \ Название транслятора
litb 06 b, PTYPE                 \ Вывести на индикатор
testAAA                          \ Примитивный диалог
BYE                              \ Выход в калькулятор

FORTH
.( Compiled: PC=) pc .  .( BC=) bc .

FRCHK

\ Запишем Каллисто-2 на диск
ONLY FORTH ALSO VMETA
save.mkp save.mkb save.mkd

.( Saved!) CR

BYE
\EOF
